package main

import (
	"bufio"
	"bytes"
	"errors"
	"fmt"
	"io"
	"os"
	"os/exec"
	"path/filepath"
	"regexp"

	"github.com/emicklei/proto"
	"github.com/sirupsen/logrus"
	"github.com/spf13/cobra"

	"gitlab.bertha.cloud/partitio/lab/protoc-gen-unannotated/protofmt"
)

var (
	extraImportRegexp = regexp.MustCompile(`^(.+):\d*:\d*: warning: Import (.*) is unused.$`)
)

func main() {
	var out string
	var imports []string
	cmd := cobra.Command{
		Use:   "protoc-gen-unannotated [protofiles ...]",
		Short: "Generate Unannotated protobuf files",
		Args:  cobra.MinimumNArgs(1),
		Run: func(cmd *cobra.Command, args []string) {
			c, _ := cmd.Flags().GetBool("clean-imports")
			for _, in := range args {
				if err := unannotate(in, out, c, imports...); err != nil {
					logrus.Fatal(err)
				}
			}
		},
	}
	cmd.Flags().StringVarP(&out, "out", "o", "", "The output directory")
	cmd.Flags().StringSliceVarP(&imports, "imports", "i", nil, "The imports used by protoc")
	cmd.Flags().BoolP("clean-imports", "c", false, "Clean unused imports (use protoc, imports may be required")
	cmd.Execute()
}

func unannotate(in string, out string, cleanImports bool, imports ...string) error {
	f, err := os.Open(in)
	if err != nil {
		return err
	}
	defer f.Close()
	w, err := makeOut(out, in)
	if err != nil {
		return err
	}
	defer w.Close()
	if !cleanImports {
		return format(f, w, nil)
	}
	var unused []string
	return withTmp(in, func(path string, tmpw io.Writer) error {
		if err := format(f, tmpw, nil); err != nil {
			return err
		}
		var is []string
		for _, i := range imports {
			is = append(is, "-I", i)
		}
		ptmp := filepath.Join(os.TempDir(), "proto")
		cmd := exec.Command("protoc", append(is, "-o", ptmp, "-I", filepath.Dir(path), path)...)
		defer os.Remove(ptmp)
		stderr := &bytes.Buffer{}
		cmd.Stderr = stderr
		if err := cmd.Run(); err != nil {
			return errors.New(stderr.String())
		}
		s := bufio.NewScanner(stderr)
		for s.Scan() {
			if matches := extraImportRegexp.FindStringSubmatch(s.Text()); len(matches) > 2 && matches[1] == filepath.Base(in) {
				unused = append(unused, matches[2])
			}
		}
		tmpr, err := os.Open(path)
		if err != nil {
			return err
		}
		return format(tmpr, w, unused)
	})
}

func format(f io.Reader, w io.Writer, imports []string) error {
	parser := proto.NewParser(f)
	def, err := parser.Parse()
	if err != nil {
		return err
	}
	pfmt := protofmt.NewFormatter(w, "    ")
	for _, each := range def.Elements {
		if i, ok := each.(*proto.Import); ok && contains(imports, i.Filename) {
			continue
		}
		if removeOptions(each) {
			each.Accept(pfmt)
		}
	}
	return nil
}

func contains(ss []string, s string) bool {
	for _, v := range ss {
		if v == s {
			return true
		}
	}
	return false
}

func withTmp(in string, fn func(path string, writer io.Writer) error) error {
	tmpPath := filepath.Join(os.TempDir(), filepath.Base(in))
	tmp, err := os.Create(tmpPath)
	if err != nil {
		return err
	}
	defer tmp.Close()
	defer os.Remove(tmpPath)
	return fn(tmpPath, tmp)
}

func removeOptions(v proto.Visitee) bool {
	switch e := v.(type) {
	case *proto.Option:
		return false
	case *proto.Service:
		e.Elements = removeChildOptions(e.Elements)
	case *proto.Message:
		e.Elements = removeChildOptions(e.Elements)
	case *proto.RPC:
		e.Elements = removeChildOptions(e.Elements)
	case *proto.Oneof:
		e.Elements = removeChildOptions(e.Elements)
	case *proto.OneOfField:
		e.Options = nil
		e.Field.Options = nil
	case *proto.NormalField:
		e.Options = nil
		e.Field.Options = nil
	}

	return true
}

func removeChildOptions(es []proto.Visitee) []proto.Visitee {
	var out []proto.Visitee
	for _, v := range es {
		if removeOptions(v) {
			out = append(out, v)
		}
	}
	return out
}

func makeOut(out, in string) (io.WriteCloser, error) {
	if out == "" {
		return &stdout{}, nil
	}
	out = filepath.Join(out, in)
	if err := os.MkdirAll(filepath.Dir(out), os.ModePerm); err != nil {
		return nil, err
	}
	w, err := os.Create(out)
	if err != nil {
		return nil, err
	}
	return w, nil
}

type stdout struct {
}

func (s *stdout) Write(p []byte) (n int, err error) {
	fmt.Print(string(p))
	return len(p), nil
}

func (*stdout) Close() error {
	return nil
}
